//-------+---------+---------+---------+---------+---------+---------+--------=
//
// File: coverage_dist.c
//
//----------
//
// coverage_dist--
//	Support for collecting the query coverage distribution from alignments.
//
//----------

//----------
//
// other files
//
//----------

#include <stdlib.h>				// standard C stuff
#define  true  1
#define  false 0
#include <string.h>				// standard C string stuff
#include <ctype.h>				// standard C upper/lower stuff
#include <stdarg.h>				// standard C variable argument list stuff
#include "utilities.h"			// utility stuff
#include "dna_utilities.h"		// dna/scoring stuff
#include "sequences.h"			// sequence stuff
#include "edit_script.h"		// alignment edit script stuff

#define  coverage_dist_owner	// (make this the owner of its globals)
#include "coverage_dist.h"		// interface to this module

//----------
//
// filter_aligns_by_coverage--
//	Filter a list of alignments, removing any alignment that has query coverage
//	percentage outside of a specified range.
//
//----------
//
// Arguments:
//	seq*		seq1, seq2:		The sequences.
//	alignel*	alignList:		The list of alignments to operate upon.
//	float		minCoverage,	The range of query coverage that we will *keep*.
//				maxCoverage		.. These are values between 0 and 1.
//
// Returns:
//	A pointer to the list of remaining alignments.
//
//----------
// Notes:
//	(1)	Memory for alignments that don't make the cut is deallocated here.
//	(2)	The returned list of alignments is in the same order as the incoming
//		.. list.
//	(3) Coverage is counted over all aligned bases, included those aligned to
//		.. gaps.
//----------

alignel* filter_aligns_by_coverage
   (arg_dont_complain(seq* seq1),
	seq*		seq2,
	alignel*	alignList,
	float		minCoverage,
	float		maxCoverage)
	{
	alignel*	a, *next;
	alignel*	head, *prev;
	unspos		numer, denom;
	float		minCov, maxCov;

	denom = seq2->len;
	minCov = denom * minCoverage;
	maxCov = denom * maxCoverage;

	head = prev = NULL;
	for (a=alignList ; a!=NULL ; a=next)
		{
		next  = a->next;
		numer = a->end2+1 - a->beg2;

		if ((numer < minCov) || (numer > maxCov))
			{ // (unwanted alignment, discard it)
			free_if_valid ("filter_aligns_by_coverage a->script", a->script);
			free_if_valid ("filter_aligns_by_coverage a",         a);
			continue;
			}

		// this alignment is ok, add it to the end of the new list we're
		// building

		if (head == NULL) head = prev = a;
		             else { prev->next = a;  prev = a; }

		a->next = NULL;
		}

	return head;
	}

//----------
//
// filter_segments_by_coverage--
//	Filter a table of segments, removing any segment that has query coverage
//	percentage outside of a specified range.
//
//----------
//
// Arguments:
//	seq*		seq1, seq2:		The sequences.
//	segtable*	st:				The segment table to operate upon.
//	float		minCoverage,	The range of query coverage that we will *keep*.
//				maxCoverage		.. These are values between 0 and 1.
//
// Returns:
//	(nothing)
//
//----------

void filter_segments_by_coverage
   (arg_dont_complain(seq* seq1),
	seq*		seq2,
	segtable*	st,
	float		minCoverage,
	float		maxCoverage)
	{
	segment*	srcSeg, *dstSeg;
	unspos		numer, denom;
	float		minCov, maxCov;

	if (st      == NULL) return;
	if (st->seg == NULL) return;

	denom = seq2->len;
	minCov = denom * minCoverage;
	maxCov = denom * maxCoverage;

	for (dstSeg=srcSeg=st->seg ; ((u32)(srcSeg-st->seg))<st->len ; srcSeg++)
		{
		numer = srcSeg->length;
		if ((numer < minCov) || (numer > maxCov))
			continue; // (unwanted segment, skip it)

		if (srcSeg != dstSeg) *dstSeg = *srcSeg;
		dstSeg++;
		}

	st->len = dstSeg - st->seg;
	}

